LANGUAGE 0, 0
100 ICON  icon.ico

VS_VERSION_INFO VERSIONINFO
 FILEVERSION 0,7,0,0
 PRODUCTVERSION 0,7,0,0
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x4L
 FILETYPE 0x2L
 FILESUBTYPE  0x0L
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "040904B0"
    BEGIN
      VALUE "Comments", "xcalib calibration loader\0"
      VALUE "FileDescription", "xcalib\0"
      VALUE "LegalCopyright", "Copyright (c) 2004-2007 Stefan Doehla\0"
      VALUE "OriginalFilename", "xcalib.exe\0"
      VALUE "ProductVersion", "0,7\0"
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
    VALUE "Translation", 0x0409, 0x04B0
  END
END
